#include "compiler.h"
#include "bin2abc.h"

static uint16_t relocs[MAX_ADDR];
static int nrelocs;

/*
 * This flag is set if we have *any* relocations set or pushed,
 * including if we have read a null file!
 */
bool have_relocs;

/*
 * Load a file of relocations
 */
void read_reloc_file(const char *filename)
{
    size_t n;

    have_relocs = true;

    n = read_file(relocs + nrelocs, 2 * (MAX_ADDR - nrelocs), filename);
    n >>= 1;

#ifdef WORDS_LITTLEENDIAN
    nrelocs += n;
#else
    while (n--) {
        relocs[nrelocs] = letohs(relocs[nrelocs]);
        nrelocs++;
    }
#endif
}

/*
 * Push additional relocations onto the list
 */
void push_reloc(uint16_t r)
{
    relocs[nrelocs++] = r;
}

/*
 * Pack relocations; this is done in place.
 * The value delta is added to the first replocation skip.
 *
 * If adjust is set then the first relocation is adjusted to
 * add the size of the relocation data itself to it.
 */
static int sort_base;

/*
 * Sort addresses, keeping in mind the sort_base starting point
 * Correct wraparound behavior here is really important!
 */
static int sort_addrs(const void *p1, const void *p2)
{
    uint16_t a1 = *(const uint16_t *)p1 - sort_base;
    uint16_t a2 = *(const uint16_t *)p2 - sort_base;

    return a1 - a2;
}

int pack_relocs(uint8_t ** out, int delta, bool adjust)
{
    uint8_t *p, *rbuf;
    const uint16_t *rp;
    uint16_t r, r0;
    unsigned int rd;
    int last;
    int len;

    sort_base = -delta;
    qsort(relocs, nrelocs, sizeof *relocs, sort_addrs);

    p = rbuf = (uint8_t *) relocs + 1;  /* +1 in case we need adjust */

    rp = relocs;
    r0 = *rp + delta;		/* Save the first relocation */

    last = 0;
    while (nrelocs--) {
	r = *rp++ + delta;
	rd = r - last;
        if (rd == 0) {
            continue;           /* Duplicated relocation */
        } else if (rd < 0x80) {
            *p++ = rd;
        } else if (rd < 0x7eff) {
            *p++ = (rd >> 8) + 0x81;
            *p++ = rd;
        } else {
            fprintf(stderr,
                    "%s: impossibly large relocation jump: 0x%x bytes\n",
                    program_name, r);
            return -1;
        }
        last = r + 2;
    }
    *p++ = 0x80;                /* End markerxs */
    len = p - rbuf;

    /* If len == 1 then all we have is the end marker, nothing to adjust */
    if (len > 1 && adjust) {
	uint16_t r1 = r0 + len;

        if (r1 > 0x7f) {
            if (r0 <= 0x7f) {
                rbuf--;         /* We saved a byte for this just in case */
                r1++;           /* Add one more byte to the reloc length */
                len++;
            }

            rbuf[0] = (r1 >> 8) + 0x81;
            rbuf[1] = r1;
        } else {
            rbuf[0] = r1;
        }
    }

    *out = rbuf;
    return len;
}

/*
 * Apply relocations; this is the only option when dealing with
 * absolute file formats like .abs.
 */
void apply_relocs(uint8_t *data, uint16_t origin)
{
    const uint16_t *rp;
    int i;

    data += origin;

    rp = relocs;
    for (i = 0; i < nrelocs; i++) {
	uint16_t r = *rp++;

	set16(data+r, get16(data+r) + origin);
    }
}
